''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'''                                                                    '''
''' CODE GENERATOR DEMO        *** released to the public domain ***   '''
'''                                                                    '''
''' By Rowan Crowe, released Monday 28-Dec-1998                        '''
'''   rowan@sensation.net.au                                           '''
'''   3:635/728.1@fidonet                                              '''
'''                                                                    '''
''' This program is similar to CALC.MOO except instead of just         '''
''' evaluating and reporting the answer, it generates 80x86 ASM        '''
''' code to demonstrate how it's done. An intermediate pseudo code     '''
''' is used to allow for peephole optimisation before output as        '''
''' as 80x86 ASM in text form. The (now unfinished) new version of     '''
''' MoonRock used this type of pseudo code and peephole optimisation.  '''
'''                                                                    '''
''' To see the optimisation in action use the -O switch on the         '''
''' commandline. (More things could be optimised, but this is a demo)  '''
'''                                                                    '''
''' This code has some remnants of higher level functionality like     '''
''' less than/equal to/greater than evaluation and function calls.     '''
''' Try "fred(a+b, c-3)" to see the output. :)                         '''
'''                                                                    '''
''' Requires MoonRock compiler v0.50 or higher:                        '''
'''   http://www.rowan.sensation.net.au/moonrock.html                  '''
'''                                                                    '''
'''   MRC codegen                                                      '''
'''         or                                                         '''
'''   MRC codegen !DEBUG                                               '''
'''                                                                    '''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

begin def

stack 8k

%BPLIMIT = 128        ' max number of discrete elements: convenient number
dim bp$[%BPLIMIT]
dim bpt%[%BPLIMIT]

%PEEPHOLE_LIMIT = 256
dim peephole_op%[%PEEPHOLE_LIMIT]
dim peephole_reg1%[%PEEPHOLE_LIMIT]
dim peephole_reg2%[%PEEPHOLE_LIMIT]
dim peephole_parm1$[%PEEPHOLE_LIMIT]
dim peephole_parm2$[%PEEPHOLE_LIMIT]

%OPERATOR = 1         ' + - / *
%ALPHA = 2            ' letter
%BRACKET = 3          ' (
%COMMA = 4
%FUNCTION = 5         ' function()
%NUMBER = 6           ' digit 0-9

set = 0,   %OP_NULL, %OP_MOV, %OP_PUSH, %OP_POP, %OP_ADD, %OP_SUB, _
           %OP_DIV, %OP_MUL, %OP_CALL, %OP_CMP, %OP_SBB, %OP_JG, %OP_JL, _
           %OP_INC, %OP_LABEL
set = 1,   %IMM, %VAR, %AX, %BX, %SP

common bpp%, bpptr%, negflag%, peephole_ptr%, locallabel%

bundle switch optimise%

sub Parse(t$)
sub Evaluate(t$)
sub Compare
sub AddSubtract
sub DivideMultiply
sub Bracket

sub wp(op%, reg1%, reg2%, parm1$, parm2$)
sub writeout
sub writereg(p%, p$)
sub optimise
sub peephole_bunch

strsegsize 64k

begin code

cmd$ = cmdline : cmd$ = ltrim(cmd$) : cmd$ = rtrim(cmd$)
cmd$ = ucase(cmd$)
if cmd$ = "-O" then switch.optimise% = TRUE

print "enter expression: "
input(exp$, 200)
print "\n"
call Evaluate(exp$)
' output is now sitting in pseucode form in peephole_* arrays

if switch.optimise% = TRUE then call optimise
' optimisation (if desired) changes and compacts some of the pseudocode

call writeout
' now print out the pseudocode in ASM format

#ifdef DEBUG
  s% = unusedstack
  print "unused stack space=" + s% + "\n"
#endif
end


sub Evaluate(t$)
peephole_ptr% = 0
for i% = 0 to %PEEPHOLE_LIMIT
  peephole_op%[i%] = %OP_NULL
next
call Parse(t$)
#ifdef DEBUG
  for i% = 0 to bpp%
    print bp$[i%] + "["
    select case bpt%[i%]
      case %NUMBER
        print "NUMBER"
      case %OPERATOR
        print "OPERATOR"
      case %ALPHA
        print "ALPHA"
      case %BRACKET
        print "BRACKET"
      case %COMMA
        print "COMMA"
      case %FUNCTION
        print "FUNCTION"
      case else
        print "unknown:" + bpt%[i%]
    end select
    print "]"
  next
  print "\n"
#endif

bpptr% = 0
peephole_ptr% = 0

while bpptr% <= bpp%
  call Compare
wend                             
end sub

#preserve
sub Compare
' do the contents of this sub actually get used? RC 1998/12/28
call AddSubtract
if bpptr% > bpp% then exit sub
r% = sinstr(bp$[bpptr%], "<=>")
while r% <> 0
  select case bp$[bpptr%]
    case "="
      bpptr% = bpptr% + 1
      call wp(%OP_PUSH, %AX, 0, "", "")
      call DivideMultiply
      call wp(%OP_POP, %BX, 0, "", "")
      call wp(%OP_CMP, %AX, %BX, "", "")
      call wp(%OP_MOV, %AX, %IMM, "", "0")
      call wp(%OP_SBB, %AX, %IMM, "", "0")
    case "<"
      call wp(%OP_PUSH, %AX, 0, "", "")
      call DivideMultiply
      call wp(%OP_POP, %BX, 0, "", "")
      call wp(%OP_CMP, %AX, %BX, "", "")
      call wp(%OP_MOV, %AX, %IMM, "", "0ffffh")
      t$ = "LL" + locallabel% : locallabel% = locallabel% + 1
      call wp(%OP_JL, %IMM, 0, t$, "")
      call wp(%OP_INC, %AX, 0, "", "")
      call wp(%OP_LABEL, %IMM, 0, t$, "")
    case ">"
      call wp(%OP_PUSH, %AX, 0, "", "")
      call DivideMultiply
      call wp(%OP_POP, %BX, 0, "", "")
      call wp(%OP_CMP, %AX, %BX, "", "")
      call wp(%OP_MOV, %AX, %IMM, "", "0ffffh")
      t$ = "LL" + locallabel% : locallabel% = locallabel% + 1
      call wp(%OP_JG, %IMM, 0, t$, "")
      call wp(%OP_INC, %AX, 0, "", "")
      call wp(%OP_LABEL, %IMM, 0, t$, "")

  end select
  r% = sinstr(bp$[bpptr%], "<=>")
wend
end sub


#preserve
sub AddSubtract
call DivideMultiply
if bpptr% > bpp% then
  exit sub
endif
r% = sinstr(bp$[bpptr%], "+-")
while r% <> 0
  if bp$[bpptr%] = "+" then
    bpptr% = bpptr% + 1
    call wp(%OP_PUSH, %AX, 0, "", "")
    call DivideMultiply
    call wp(%OP_POP, %BX, 0, "", "")
    call wp(%OP_ADD, %AX, %BX, "", "")
  else
    bpptr% = bpptr% + 1
    call wp(%OP_PUSH, %AX, 0, "", "")
    call DivideMultiply
    call wp(%OP_MOV, %BX, %AX, "", "")
    call wp(%OP_POP, %AX, 0, "", "")
    call wp(%OP_SUB, %AX, %BX, "", "")
  endif
  r% = sinstr(bp$[bpptr%], "+-")
wend

end sub


#preserve
sub DivideMultiply
call Bracket
if bpptr% > bpp% then
  exit sub
endif
r% = sinstr(bp$[bpptr%], "*/")
while r% <> 0
  if bp$[bpptr%] = "*" then
    bpptr% = bpptr% + 1
    call wp(%OP_PUSH, %AX, 0, "", "")
    call Bracket
    call wp(%OP_MOV, %BX, %AX, "", "")
    call wp(%OP_POP, %AX, 0, "", "")
    call wp(%OP_MUL, %BX, 0, "", "")
  else
    bpptr% = bpptr% + 1
    call wp(%OP_PUSH, %AX, 0, "", "")
    call Bracket
    call wp(%OP_MOV, %BX, %AX, "", "")
    call wp(%OP_POP, %AX, 0, "", "")
    call wp(%OP_DIV, %BX, 0, "", "")

  endif
  r% = sinstr(bp$[bpptr%], "*/")
wend
end sub


#preserve
sub Bracket
if bpptr% > bpp% then exit sub

negjump:

select case bpt%[bpptr%]
  case %NUMBER
    t$ = bp$[bpptr%]
    if negflag% = TRUE then
      t$ = "-" + t$
      negflag% = FALSE
    endif
    call wp(%OP_MOV, %AX, %IMM, "", t$)
    bpptr% = bpptr% + 1
  case %FUNCTION
    fn% = bpptr%
    bpptr% = bpptr% + 1
    if bp$[bpptr%] <> "(" then
      print "syntax error: '(' expected after function name ('" + bp$[bpptr%] + "' found)\n"
      end
    endif
    bpptr% = bpptr% + 1
    sp% = 0
    if bp$[bpptr%] <> ")" then
      found% = FALSE
      while found% = FALSE
        call Compare
        if bp$[bpptr%] = ")" then found% = TRUE : exit while
        if bp$[bpptr%] = "," then sp% = sp% + 2
        if bpptr% > bpp% then exit while
      wend

      call wp(%OP_PUSH, %AX, 0, "", "")
      sp% = sp% + 2
    endif

    t$ = "_" + bp$[fn%]
    call wp(%OP_CALL, %IMM, 0, t$, "")
    if sp% <> 0 then
      t$ = str(sp%)
      call wp(%OP_ADD, %SP, %IMM, "", t$)
    endif
    bpptr% = bpptr% + 1
  case %ALPHA
    call wp(%OP_MOV, %AX, %VAR, "", bp$[bpptr%])
    bpptr% = bpptr% + 1
  case %COMMA
    call wp(%OP_PUSH, %AX, 0, "", "")
    bpptr% = bpptr% + 1
  case %BRACKET
    bpptr% = bpptr% + 1
    call Compare
    if bp$[bpptr%] <> ")" then
      print "syntax error: expected ')'\n"
      end
    endif
    bpptr% = bpptr% + 1
  case else
    ' handle negative numbers
    if bp$[bpptr%] = "-" then
      if negflag% = FALSE then
        bpptr% = bpptr% + 1
        negflag% = TRUE
        goto negjump
      else
        ' too many '-'s
        print "syntax error: unexpected token '" + bp$[bpptr%] + "' (too many negates)\n"
       end
      endif
    else
      print "syntax error: unexpected token '" + bp$[bpptr%] + "'\n"
      end
    endif
end select

end sub


sub Parse(t$)
bpp% = 0

if t$ = "" then exit sub
t$ = t$ + ")"
p$ = ""
lt% = len(t$)
for i% = 1 to lt%
  c% = sbyte(t$, i%)
'  print i% + ":" + chr(c%) + "\n"
  select case c%
    case 43, 45, 47, 42, 60, 61, 62, 40, 41, 44 ' +-/*<=>(),
      if p$ <> "" then
        p$ = lcase(p$)
        bp$[bpp%] = p$
        t% = asc(p$)
        p$ = ""
        if t% > 95 then
          if t% < 123 then
            bpt%[bpp%] = %ALPHA
            l% = len(bp$[bpp%])
            if l% > 1 then bpt%[bpp%] = %FUNCTION
          endif
        endif
        if t% > 47 then
          if t% < 58 then
            bpt%[bpp%] = %NUMBER
          endif
        endif

        bpp% = bpp% + 1
      endif

      bp$[bpp%] = chr(c%)
      t% = sbyte(t$, i%)

      select case t%
        case 43, 45, 47, 42
          bpt%[bpp%] = %OPERATOR
        case 60  ' <
          bpt%[bpp%] = %OPERATOR
        case 61  ' =
          bpt%[bpp%] = %OPERATOR
          bpp_1% = bpp% - 1
          if bp$[bpp_1%] = "<" then
            bpp% = bpp% - 1
            bp$[bpp%] = "<="
          endif
        case 62  ' >
          bpt%[bpp%] = %OPERATOR
          bpp_1% = bpp% - 1
          if bp$[bpp_1%] = "=" then
            bpp% = bpp% - 1
            bp$[bpp%] = "=>"
          endif
        case 40, 41
          bpt%[bpp%] = %BRACKET
        case 44
          bpt%[bpp%] = %COMMA
      end select
      bpp% = bpp% + 1
    case 32
     ' nothing
    case else
      p$ = p$ + chr(c%)
  end select
next
print "\n"

bpp% = bpp% - 1
bp$[bpp%] = "<EOL>"
bpt%[bpp%] = 0
bpp% = bpp% - 1

end sub


sub wp(op%, reg1%, reg2%, parm1$, parm2$)
peephole_op%[peephole_ptr%] = op%
peephole_reg1%[peephole_ptr%] = reg1%
peephole_reg2%[peephole_ptr%] = reg2%
peephole_parm1$[peephole_ptr%] = parm1$
peephole_parm2$[peephole_ptr%] = parm2$
if peephole_ptr% = %PEEPHOLE_LIMIT then
  print "fatal: peephole buffer limit exceeded.\n"
  end
endif
peephole_ptr% = peephole_ptr% + 1
end sub


sub writeout
for i% = 0 to peephole_ptr%
  select case peephole_op%[i%]
    case %OP_MOV
      print "\tmov\t"
      call writereg(peephole_reg1%[i%], peephole_parm1$[i%])
      print ","
      call writereg(peephole_reg2%[i%], peephole_parm2$[i%])
      print "\n"
    case %OP_PUSH
      print "\tpush\t"
      call writereg(peephole_reg1%[i%], "")
      print "\n"
    case %OP_POP
      print "\tpop\t"
      call writereg(peephole_reg1%[i%], "")
      print "\n"
    case %OP_ADD
      print "\tadd\t"
      call writereg(peephole_reg1%[i%], peephole_parm1$[i%])
      print ","
      call writereg(peephole_reg2%[i%], peephole_parm2$[i%])
      print "\n"
    case %OP_SUB
      print "\tsub\t"
      call writereg(peephole_reg1%[i%], peephole_parm1$[i%])
      print ","
      call writereg(peephole_reg2%[i%], peephole_parm2$[i%])
      print "\n"
    case %OP_DIV
      print "\tdiv\t"
      call writereg(peephole_reg1%[i%], peephole_parm1$[i%])
      print "\n"
    case %OP_MUL
      print "\tdiv\t"
      call writereg(peephole_reg1%[i%], peephole_parm1$[i%])
      print "\n"
    case %OP_CALL
      print "\tcall\t" + peephole_parm1$[i%] + "\n"
    case %OP_CMP
      print "\tcmp\t"
      call writereg(peephole_reg1%[i%], peephole_parm1$[i%])
      print ","
      call writereg(peephole_reg2%[i%], peephole_parm2$[i%])
      print "\n"
    case %OP_SBB
      print "\tsbb\t"
      call writereg(peephole_reg1%[i%], peephole_parm1$[i%])
      print ","
      call writereg(peephole_reg2%[i%], peephole_parm2$[i%])
      print "\n"
    case %OP_NULL
    ' nothing
'    print "null\n"
    case else
      print "unknown opcode type 0x" + hex16(peephole_op%[i%]) + " at peephole position " + i% + "\n"
  end select
next
end sub


sub writereg(p%, p$)
select case p%
  case %IMM
    print p$
  case %VAR
    print p$
  case %AX
    print "ax"
  case %BX
    print "bx"
  case %SP
    print "sp"
end select
end sub


sub optimise
for i% = 0 to peephole_ptr%

  ' optimise:
  '   push  ax
  '   mov   ax,blah
  '   pop   bx
  '   add   ax,bx
  ' to:
  '   mov   bx,blah
  '   add   ax,bx

  if peephole_op%[i%] = %OP_PUSH then
    if peephole_reg1%[i%] = %AX then
      i1% = i% + 1
      if peephole_op%[i1%] = %OP_MOV then
        if peephole_reg1%[i1%] = %AX then
          i2% = i% + 2
          if peephole_op%[i2%] = %OP_POP then
            if peephole_reg1%[i2%] = %BX then
              i3% = i% + 3
              if peephole_reg1%[i3%] = %AX then
                if peephole_reg2%[i3%] = %BX then
                  if peephole_op%[i3%] = %OP_ADD then
                    peephole_op%[i%] = %OP_NULL
                    peephole_reg1%[i1%] = %BX
                    peephole_op%[i2%] = %OP_NULL
                    call peephole_bunch
                  endif
                endif
              endif
            endif
          endif
        endif
      endif
    endif
  endif

' optimise:
'   mov   bx,blah
'   add   ax,bx
' to:
'   add   ax,blah
  if peephole_op%[i%] = %OP_MOV then
    if peephole_reg1%[i%] = %BX then
      i1% = i% + 1
      if peephole_op%[i1%] = %OP_ADD then
        if peephole_reg1%[i1%] = %AX then
          if peephole_reg2%[i1%] = %BX then
            peephole_op%[i%] = %OP_NULL
            peephole_reg2%[i1%] = peephole_reg2%[i%]
            peephole_parm2$[i1%] = peephole_parm2$[i%]
            call peephole_bunch
          endif
        endif
      endif
    endif
  endif

' optimise:
'   mov   ax,1
'   add   ax,2
' to:
'   mov   ax,3
  if peephole_op%[i%] = %OP_MOV then
    if peephole_reg1%[i%] = %AX then
      if peephole_reg2%[i%] = %IMM then
        i1% = i% + 1
        if peephole_op%[i1%] = %OP_ADD then
          if peephole_reg1%[i1%] = %AX then
            if peephole_reg1%[i1%] = %IMM then
              t1% = val(peephole_parm2$[i%])
              t2% = val(peephole_parm2$[i1%])
              t1% = t1% + t2%
              peephole_parm2$[i%] = str(t1%)
              peephole_op%[i1%] = %OP_NULL
              call peephole_bunch
            endif
          endif
        endif
      endif
    endif
  endif

' optimise:
'   push  ax
'   mov   ax,blah
'   mov   bx,ax
'   pop   ax
'   sub   ax,bx
' to:
'   mov   bx,blah
'   sub   ax,bx

next
end sub


sub peephole_bunch
'print "bunch\n"
for i% = 0 to peephole_ptr%
  if peephole_op%[i%] = %OP_NULL then
    k% = i% + 1
    for j% = i% to peephole_ptr%
'      print j% + " "
      peephole_op%[j%] = peephole_op%[k%]
      peephole_reg1%[j%] = peephole_reg1%[k%]
      peephole_reg2%[j%] = peephole_reg2%[k%]
      peephole_parm1$[j%] = peephole_parm1$[k%]
      peephole_parm2$[j%] = peephole_parm2$[k%]
      k% = k% + 1
    next
    peephole_ptr% = peephole_ptr% - 1
'    print "\n"
  endif
next
end sub

